;(c) Wally Bass 1995, 2008. Released to the public domain by author
;
;This include file is designed support programs which do simple stream
;oriented conversions on files, with one input file and one output file,
;such as CR-LF to LF conversion or visa versa. The code is organized
;around the idea of supporting a tight loop for the actual data
;examination and processing, where DS:SI tracks the input stream and
;ES:DI tracks the output stream (see 'mainloop'). One segment (DS) is
;devoted to the input file, one (ES) to the output file. Code, stack,
;and all variables outside of the buffers are held in a third segment
;(CS, SS). The unit of disk input and output is nominally 0F800h bytes,
;but is adjusted downward at startup if there is not enough memory to
;support buffers of this size.
;
;The main loop should be written to accomodate the notions that data is
;obtained from an input stream tracked by DS:SI, data is moved to an
;output stream tracked by ES:DI, and CX is maintained within the main
;loop to indicate the lesser of remaining input bytes currently
;available (following DS:SI) or the remaining space (following ES:DI)
;available in the output buffer. As bytes or "chunks" of data in the
;input stream are examined and processed, and corresponding output bytes
;are passed to the output stream, DS:SI and ES:DI are advanced
;accordingly, and for each byte or "packet" processed, CX is decremented
;by the greater of the number of input bytes or the number of output
;bytes associated with the packet just processed. For example, if the
;program were a program to translate "DOS" files to "Unix" files by
;converting CR-LF character pairs to just LF chararacters pairs, most
;"chunks" processed in the file would be just a single byte (with CX
;decremented by 1), except for input CR-LFs encountered, where the input
;"chunk" would be 2 bytes (CR-LF), the output packet 1 byte (LF), and CX
;would be decremented by 2. This approach to decrementing CX means that
;CX might at any instant might be less than EITHER the input bytes or
;output space remaining, but it insures that neither the input nor
;output buffer are overrun.
;
;As CX decrements and approaches zero in the main loop, a procedure
;called "stdrw" is called to do actual IO, to refresh the input stream
;buffer with more data, and to write output buffer data so as to create
;more available space for output. This call will then set CX to the
;lesser of the input bytes available (in the input buffer) or space
;available (in the out buffer). There is a MINBUF assembly time variable
;that should be set before including this include file -- setting this
;variable insures that CX will be MINBUF*16 or more after any call to
;"stdrw," except for when processing is near the end of the input file
;and there are less than MINBUF*16 remaining to be processed. MINBUF can
;be set anywhere in the range 1 to 7Fh. Using a higher then necessary
;MINBUF value causes extra and unnecessary data movement to occur within
;the "stdrw" routine, and may result in data being read/written to the
;input/output files in smaller amounts, so you should use the smallest
;MINBUF value that works for what you are doing.
;
;All of this means that a routine that a program that simply passes all
;input data to the output would require only the code
;
;	rdwr:	call	stdwr
;		jcxz	done
;	mainloop:lodsb
;		stosb
;		loop	mainloop
;		jmp	rdwr
;
;for its main processing loop. (In this trivial case, the
;lodsb,stosb,loop sequence could even be replaced by a "rep movsb"
;instruction, but in the general case, the lodsb code is used because
;the input is being examined in some way.)
;
;A complete program that copys a file, but, say, converts CR-LF pairs to
;LF's and converts single quotes to a pair of single quotes (to show
;examples of both contraction and expansion) could be constructed as
;follows:
;
DOEXAMPLE = 0
IF DOEXAMPLE
;*********************************************************************
;**************************		    **************************
;************************** example program **************************
;**************************		    **************************
;*********************************************************************
	page	,132	;or whatever
	title	whatever
HOOKS = 1	;set to 1 to put in debugging "hooks."
		;search for "HOOKS" to see more discussion
;
MINBUF = 1	;One paragraph (16 bytes) is plenty. We only need two
		;bytes target space (for CR-LF pair)
;
;The 'inout.inc' include file executes the macro 'doparm1' before
;parsing the the two filename parms, executes the macro 'doparm2' after
;parsing the the two filename parms, and executes the macro 'doparm3' in
;a place where you can insert a parm parsing subroutine or two. If your
;app has more than the two filename parms, you might include some
;content in these macros to parse the parms. Otherwise, you must provide
;empty macros by these names, as done here.
;
;Keep in mind that any code defined here is overlayed by the stack, once
;the input and output file have been opened. So code in your "main loop"
;cannot refer to code defined here. If you use these facilities, you
;should inspect the code at doparm1 and doparm2 to get an idea of
;register usage -- generally, DX and DI must be preserved and AH must
;set back to zero if used.
;
doparm1 macro
	endm
doparm2 macro
	endm
doparm3 macro
	endm
;
;'inout.inc' also invokes the macros "doread" right after a successful
;read to the input file, and "dowrite" before each write to the output
;file. You can define code in these macros if you need to inspect or
;modifiy the input or outpu stream in some way that is not all that
;synchronized with what is happening in the main loop. Be sure to save
;all registers and leave the direction flag cleared if you insert code
;at these points.
;
doread	macro
	endm
dowrite macro
	endm
;
;The 'inout.inc' include file executes the macro 'dovar' at an
;appropriate place for allocating variables. If you define your
;variables this way, they will be placed with the variables defined
;by/for the inout.inc itself. Word alignment exists when the macro is
;executed
;
dovars	macro
wascr	db	0	;unexpected CR flag
waslf	db	0	;unexpected LF flag
	endm
;
;The inout.inc file should be included immediately, and should be
;allowed to define the code segment. Code owned by 'inout.inc' must be
;first in the module... after parameters are parsed, the parameter
;parsing code area is used as the stack area. inout.inc defines the
;segment ('_TEXT') and generates code to parse parms and open the input
;and output files, including suitable error handling.
;
include inout.inc
;
;The main loop can be included next. The code in 'inout.inc' concludes
;by jumping to the label 'rdwr', and the input buffer is empty at this
;point so you must do a call to the I/O routines before you start
;processing. The label 'done' is above this point, and a
;short/conditional jump to 'done' may work depending on how much code
;proceeds the jump.
;
;In the main loop, 'assume cs:_TEXT,ss:_TEXT,ds:nothing,es:nothing' is
;in effect. CS and SS both point at the _TEXT segment.
;
rdwr:	call	stdrw	;sets CX=min(in_data_avail,out_space_avail),
			;DX=CX, AX=0, and BX=0
	jcxz	done	;CX==0 after stdrw implies EOF on input
;
;AX, BX, and BP are the only registers available for your use in the
;main loop, unless you save some regs. If you call rfshchk, BX, CX, and
;DX are all parameters. You provide the BX parameter.
;
;Note that AX and BX are cleared by the call to stdrw. I have found that
;code in the main loop can often take advantage of that, especially if
;one is careful not to touch AH and BH.
;
mainloop:lodsb		;input byte to AL
	cmp	al,13	;check for CR
	je	iscr
	cmp	al,10	;check for LF
	je	islf
	cmp	al,"'"	;check for quote
	je	isquote
put1:	stosb		;AL to output steam
	loop	mainloop
	jmp	short rdwr
;
;code to handle the various special cases follows
;
iscr:	mov	bl,2	;need at least two bytes of input before we can
			;even check for CR-LF pair (BH was already 0)
	call	rfshchk ;insure space is available (BX, CX, and DX are
			;all parms to 'rfshchk'
	jc	rdwr	;carry: count is low (<BX), but may be out of
			;date (e.g., it is less than DX). Call
			;read/write routines to insure CX is up to date
	jz	iscr02	;on last byte of file, unexpected CR
	cmp	byte ptr ds:[si],10	;check for LF following CR
	jne	iscr02	;jmp if not, unexpected CR
	lodsb		;CR-LF found, skip CR and get the LF
	dec	cx	;account for the byte just gotten. We are
			;assured that this will not make CX underflow
			;(go zero or negative) because of the call to
			;rfshchk with BX=2
	jmp	short put1	;return to main loop to put the LF and
				;account for it
;
iscr02: mov	cs:wascr,al	;unexpected CR (no trailing LF)
	jmp	short put1	;handle as any char after setting flag
;
islf:	mov	cs:waslf,al	;unexpected LF (no preceeding CR)
	jmp	short put1	;handle as any char after setting flag
;
isquote:mov	bl,2	;need at least 2 bytes of output space
	call	rfshchk ;insure space is available (BX, CX, and DX are
			;all parms to 'rfshchk'
	jc	rdwr	;carry: count is low (<BX), but may be out of
			;date (e.g., it is less than DX). Call
			;read/write routines to insure CX is up to date.
	sub	cx,bx	;'rfshchk' returns a safe and correct value in
			;BX for decrementing CX in this situation (data
			;expansion, more output generated than input
			;processed)
	stosb		;add quote to output stream and advance dest ptr
	jmp	short put1	;let the loop put the the other quote
;
;This completes the main loop. The 'rfshchk' routine is critical and not
;quite as simple as it may seem, so be sure to read carefully the
;description of that routine.
;
;A few additional items are required to
;complete the program:
;
;1. The terminating code in the inout.inc module calls a procedure
;called 'warnings,' which may be structured like the one below. This
;procedure can in turn use the procedure 'emsg' to display messages.
;These interfaces are structured as follows:
;
;warnings:	CX=0 on entry
;		if CX is non-zero on return, the program pauses pending
;		an acknowledgement keystroke
;emsg:		Displays the '$' terminated message located by DS:DX,
;		and sets CX to 1, but only if entered with CH set to a
;		non-zero value
;
;when control is passed to these terminal routines, DS points to the
;_TEXT segment, as well as CS and SS.
;
	assume	ds:_TEXT
warnings proc	near
	xchg	wascr,ch
	mov	dx,offset crmsg
	call	emsg
	xchg	waslf,ch
	mov	dx,offset lfmsg
	call	emsg
	ret
warnings endp
;
;2. A usage message for the program, with the label 'usemsg'
;
usemsg	db	'Usage: CRLF2LF infile outfile',13,10,'$'
;
;3. Any error message that you want to define
;
crmsg	db	7,"WARNING: input had CR's w/o LF's",13,10,'$'
lfmsg	db	7,"WARNING: input had LF's w/o CR's",13,10,'$'
;
;4. An ending for the module, consisting of the following three
;statements:
;
IF HOOKS
;If you set the HOOKS assembly time variable to 1, then the very first
;thing done after the program starts is to call a routine called
;"debug1". And a call to "debug2" is done right before before control is
;passed to the label "rdwr" in your program. So, if HOOKS is set, YOU
;must provide two procs by those names. We suggest the following
;structure.
	db	'start'
debug1	proc	near
	nop
	ret
debug1	endp
	db	'mainlp'
debug2	proc	near
	nop
	ret
debug2	endp
;
;The idea is that you will easily be able to find these routines when
;you run your program under a debugger, because you only have to single
;step one instruction into the program to arrive at "debug1". The string
;constants placed in memory here then serve as a reminder as to when
;these routines are called. Our main loop in this example wasn't
;interesting enough to have any more debug points like this, but you
;could have several more such routines, each preceed by a string to
;remind you when it is called. They should all be grouped together --
;the whole idea is to be able to find the group of routines by the early
;call to "debug1", and then find the rest via in memory via the text
;strings, the "nop, ret" code structure of each routine, and the fact
;that all such routines are in close proximity.
;
;You can then set breakpoints on whichever NOPs you want, or change the
;NOPs to "int 3"s, with your debugger, to quickly get to various debug
;points that you might be interested in.
endIF
	dobuffer	;defines the buffer area origin
_TEXT	ends
	end	start
;*********************************************************************
;**************************		    **************************
;**************************  end of example **************************
;**************************		    **************************
;*********************************************************************
endIF
;
dobuffer macro
t = (offset s0+8000h-$) mod 16
IF t
	db	t dup(0)	;to paragraph boundary for buffer
endIF
;
buffer	equ	this word
t = (offset s0+8100h-$) mod 512
IF t
	db	t dup(0)	;pad to mult of 512 for faster loading
endIF
	endm
;
_TEXT	segment WORD PUBLIC 'CODE'
s0	equ	this byte
	assume	cs:_TEXT,ds:_TEXT,ss:_TEXT
	org	100h
;
;insure exactly two args with no wildcards. locate and null terminate
;both args
;
start	label	near
IF HOOKS
	call	debug1
endIF
	push	ds
	pop	es
	cld
	xor	ax,ax
	xor	bx,bx
	mov	si,81h		;point at parm
	mov	bl,ds:[si-1]	;get length
	mov	ds:[si+bx],bh	;terminate arg string
	call	blnkskp 	;move to first filename
	doparm1 		;maybe parse some parms here
	mov	di,si		;save start of first filename
	call	doname		;validate, null terminate filename
	jc	parmbad 	;jump if no filename
	jnz	wldcard 	;jump if wildcards
	call	blnkskp 	;move to second filename
	mov	dx,si		;save start of second filename
	call	doname		;validate, null terminate filename
	jc	parmbad 	;jump if no filename
	jnz	wldcard 	;jump if wildcards
	call	blnkskp 	;skip to next parm
	doparm2 		;maybe parse some parms here
	cmp	byte ptr ds:[si],ah	;insure no extra parm
	jne	parmbad
	jmp	parmok
;
parmbad:mov	dx,offset usemsg
	push	dx
	jmp	errexit
;
wldcard:call	errexit
	db	7,'Filename wildcards not allowed',13,10,'$'
;
;advance SI to next non-whitespace
;
blnkskp proc	near
	lodsb
	cmp	al,' '
	je	blnkskp
	cmp	al,9		;tab
	je	blnkskp
	dec	si
	ret
blnkskp endp
;
;check that parm exists, has no wildcard, and null terminate the parm at
;next whitespace.
;returns: carry: no parm. no carry, non-zero: wildcards present
;
doname	proc	near
	mov	bx,-1	;for wild card tests
	cmp	byte ptr ds:[si],1
	jb	dona10
dona02: lodsb
	cmp	al,'*'
	je	dona04
	cmp	al,'?'
	jne	dona06
dona04: inc	bx
dona06: cmp	al,' '
	je	dona08
	cmp	al,9
	je	dona08
	cmp	al,1
	jnb	dona02
dona08: mov	ds:[si-1],ah
	sbb	si,0		;clear carry, adjust SI if on null
dona10: inc	bx		;set non-zero if wild cards present
	ret
doname	endp
;
	doparm3
;
;open input and output files. check that opens succeed, and insure that
;target is either absent or of length zero. get date of input file
;
parmok: mov	si,dx	;2nd file name ptr. 1st file ptr is in DI
	cmp	word ptr ds:[si],'.'+(0*256)
	je	open02		;jmp if std out specified
	mov	ax,3D00h	;open read
	int	21h
	jc	open02
	xchg	ax,bx
	mov	ax,4202h	;seek relative to end of file
	xor	cx,cx
	xor	dx,dx
	int	21h
	xchg	ax,cx
	mov	ah,3Eh		;close
	int	21h
	or	cx,dx
	jz	open02		;okay if length zero
	call	errexit
	db	7,'Output file already exists',13,10,'$'
;
open02: mov	dx,di
	xor	ax,ax		;standard in
	cmp	word ptr ds:[di],'.'+(0*256)
	je	open04		;jmp if standard in specified
	mov	ax,3D00h
	int	21h
	jnc	open04
	call	errexit
	db	7,'Could not open input file',13,10,'$'
;
open04: mov	ds:[infile],ax
	mov	ax,1		;standard out
	cmp	word ptr ds:[si],'.'+(0*256)
	je	open06		;jmp if standard out specified
	mov	ah,3Ch
	xor	cx,cx
	mov	dx,si
	int	21h
	jnc	open06
	call	errexit
	db	7,'Could not create output file',13,10,'$'
;
;Notes on variables used for file management:
;
;Both the input and output buffers start at the beginning of a segment.
;DS points at the input buffer segment, and ES at the output buffer
;segment, and since DS and ES are basically untouched during the main
;loop processing, there isn't any need for any variables to remember
;either segment location.
;
;The size of the portion of the buffers that are read/written is equal
;to bufmax-bufmin. The actual size of the buffers is bufmax.
;
;For the read buffer, data is read in at DS:bufmin. When the buffer
;pointer (SI) gets within bufmin bytes of the end of the buffer, data
;from the end of the buffer is copied to the front of the buffer so that
;it ends at offset bufmin, and bufmax-bufmin bytes of new data are read
;in at bufmin, so as to be contiguous with the end of the previous data.
;inbufend indicates the end of the available read data, and if inbufend
;is not equal to bufmax, it means that the last read was truncated, and
;we are at end of the input file.
;
;For the output buffer, writes of bufmax-bufmin bytes are written from
;the start of the write buffer. There is bufmin bytes of "guard" space
;at the end of the write buffer, and as soon output buffer pointer (DI)
;gets into to the guard space, bufmax-bufmin bytes are written to disk,
;and any data which overflowed into the guard space is moved to the
;beginning of the buffer.
;
	even
wrkstk	equ	this word
infile	dw	-1		;input file handle (0 if stdin)
datecx	dw	0		;date CX value for input file
datedx	dw	0		;date DX value for input file
outfile dw	-1		;output file handle (1 if stdout)
bufmin	dw	MINBUF*10h	;constant.
bufmax	dw	0		;in/out buffer size, incl guard space
inbufend dw	0		;end of data in input buffer
	dovars			;user vars emitted here, if desired
;
open06: mov	outfile,ax
	mov	sp,offset wrkstk
	mov	ax,5700h	;get file date
	mov	bx,infile
	test	bx,bx
	jz	bufalloc
	int	21h
	cld
	mov	datecx,cx
	mov	datedx,dx
	jc	rderr
;
;allocate two buffers. buffers are a multiple of 800h bytes for
;efficient I/O activity, plus MINBUF paragraphs of extra area for
;overlap processing. If there are gobs of space, uses 64K buffers
;on 64K boundaries to avoid any kind of DMA boundary handling overhead.
;
bufalloc:mov	ax,ds
	add	ax,(offset buffer - offset s0)/10h
	mov	dx,ax	;save buffer segment address
	neg	ax
	add	ax,ds:[2]	;AX=available paragraphs for buffering
	cmp	ax,4000h	;plenty of room?
	jb	bufa02		;jump if not
;>256K of storage. use buffers which don't cross 64K boundaries
	add	dx,0FFFh
	and	dx,0F000h	;round up to 64K boundary
	mov	ax,0F80h	;max length
	jmp	short bufa08
;
bufa02: shr	ax,1		;compute paragraphs available per buffer
	sub	ax,80h		;subtract minbuf area allowance
	jb	bufa06		;jmp if too low
	mov	bx,0F80h	;biggest buffer we can use
	cmp	ax,bx
	jb	bufa04
	mov	ax,bx		;limit to max size
bufa04: and	ax,bx		;make size a multiple of 80h para
	jnz	bufa08		;but must be non-zero though
bufa06: call	errexit
	db	7,'Not enough memory',13,10,'$'
;
bufa08: mov	cl,4
	mov	si,ax
	shl	si,cl		;convert to bytes
	add	si,bufmin	;set SI initial condition (read)
	mov	inbufend,si	;set inbufend initial condition (read)
	mov	bufmax,si	;set buffer definition parameter (r/w)
	xor	di,di		;set DI initial condition (write)
	add	ax,80h		;max minbuf allowance
	mov	ds,dx		;set read buffer segment
	assume	ds:nothing
	add	dx,ax
	mov	es,dx		;set write buffer segment
IF HOOKS
	call	debug2
endIF
	jmp	rdwr		;to includer's main loop
;
;'read' is called when there is less than MINBUF paragraphs of input
;data left for processing. It moves remaining data down in front of the
;buffer, and gets another buffer full of data, updating DS:SI and
;inbufend accordingly.
;Inputs:
;  infile: file handle
;  DS:SI : current read point
;  BX:	=bufmax
;  CX:	count of unprocessed read data
;  DX:	=bufmin
;Outputs:
;  DS:SI : updated
;  inbufend: end of data read (EOF implied if != bufmax)
;
rderr:	call	errexit
	db	7,'Read error on input file',13,10,'$'
;
read	proc	near
	push	es		;save ES
	push	di		;save DI
	push	ds
	pop	es		;ES to read segment
	mov	di,dx		;=bufmin
	sub	di,cx		;DI to target for movedown
	push	di		;push new output point address
	rep	movsb		;move data down
	pop	si		;new output point address to SI
	pop	di		;restore DI
	pop	es		;restore ES
	mov	ah,3Fh		;read
	mov	cx,bx		;=bufmax
	mov	bx,cs:infile	;infile
	sub	cx,dx		;DX=bufmin, CX now read count
	int	21h
	cld
	jc	rderr
	doread
	add	dx,ax		;end address of data read
	mov	cs:inbufend,dx
	ret
;
read	endp
;
;'rfshchk' subroutine
;
;the 'rfshchk' routine is crucial to successful use of this foundation
;code, and you should consider its operation very carefully.
;
;In writing a main loop to be supported by this code, one usually has
;something that looks like the following
;
;	rdwr:	call		stdwr
;		jcxz		done
;	mainloop:lodsb
;		test		AL in some way
;		jmp_on_cond_1	skip1	;(may not be necessary)
;		jmp_on_cond_2	special
;	put1:	stosb
;	skip1:	loop		mainloop
;		jmp		rdwr
;	;
;	special:mov		bl,n
;		call		rfshchk
;		jc		rdwr
;		...
;
;Generally, each pass throught the mainloop, there are two
;possibilities. The first is that examination of the datum loaded in AL
;by the lodsb will result in the determination that this byte is simply
;to be be discarded (jump to "skip1") or passed as is (fall through to
;"put1". In either of these cases, but only these two cases, the loop
;statement at skip1 (and the loop fall through which branches back to
;"rdwr") takes care of all necessary "counting". At the time of "lodsb",
;CX is at least one, indicating that at least one byte is available for
;input, and at least 1 byte location is available in the output buffer.
;
;But, once a determination is made (based on the byte value loaded) that
;either there will be more than one byte output, or that additional
;input need be examined to complete processing of the current input
;"transaction," a call to "rfshchk" is in order. Before calling
;"rfshchk", BX should be loaded with a value that indicates the maximum
;number of input or output bytes -- whichever is greater -- that might
;be involved in the current "transaction". (Since BH is nearly always
;held at zero in these loops, a "mov BL,n" normally suffices for setting
;the required BX value, and saves an instruction byte.)
;
;'rfshchk' assumes that the caller will ultimately continue by getting
;the required input bytes for the transaction, putting the required
;output bytes for the transaction, and then will effectively continue
;with code which effectively does
;
;		sub		cx,bx
;		jmp		skip1
;
;to continue the loop, while properly and safely updating the CX count.
;
;'rfshchk' has two goals: (1) to make sure that enough input data and
;output space is available for the current "transaction," and (2) to
;return an appropriate value in BX so that (a) the "sub cx,bx"
;instruction in combination with the "loop" statement at "skip1" CANNOT
;decrease CX PAST zero (which would cause the loop to run wild), but
;also so that (b) the "sub cx,bx" and subsequent "loop" do correctly
;decrement CX enough to account for BOTH any input processed, and any
;output generated.
;
;"rfshchk" assumes that it is called after the "lodsb" has been done
;(that is, with SI advanced to point to just after the byte fetched),
;but before much of anything else has been changed.
;
;One of the things that rfshchk can do is to decide that buffer reading
;or writing, or other updating, might be appropriate before proceding.
;In this case, 'rfschchk' returns with the carry latch set, and in that
;case, the caller is REQUIRED to do a call to "stdrw" and "try again,"
;as occurs above by jumping to "rdwr". When "rfshchk" returns with carry
;set, it has already taken care of decrementing SI for the caller, so
;the caller need only branch back to a point equivalent to the 'rdwr'
;label in the sample code indicated. This will cause a subsequent
;"retry" of processing which "sees" the same input data byte a second
;time, and will therefore subsequently go down the same path and cause
;'rfschchk' to be called second time. If done properly, 'rfshchk' will
;never return a 'carry' on the second call (for the same input byte).
;
;When 'rfshchk' does not set carry and hence the 'jc' after "rfshchk"
;"falls through," the caller can be assured of several things.
;
;Firstly, there is enough output space to accomodate the number of bytes
;indicated in BX when the 'rfshchk' call was made.
;
;Secondly, BX has been adjusted so that "sub cx,bx" followed by a jmp
;back to "skip1" (the "loop" instruction") is a safe way to update the
;CX count.
;
;The third thing the caller would of course LIKE to be assured of is
;that "BX bytes" (the BX value at the time of the call) is available for
;input. 'rfshchk' can guarantee this most of the time, but it cannot
;guarantee this as one gets close to the end of the input file. A 'jz'
;branch (following the 'jc' retry branch) will detect the case where the
;input amount available is less than BX (at the time of the 'rfshchk'
;call). Or, the caller can skip the 'jz' branch, and just check CX to
;see if enough input is available.
;
;Inputs:
;  CX:	normal loop counter, potentially containing a lower value than
;	necessary, since for each unit of data processed, it is
;	'pessimistically' decremented by the greater of input bytes
;	utilized or output bytes generated.
;  BX:	generally, the maximum number of bytes to be processed (greater
;	of input or output) on the current 'transaction.' We are
;	basically calling 'rfshchk' to insure that BX input bytes can be
;	processed and BX output bytes can be processed without
;	additional checking of the count. One should never call
;	'rfshchk' with a BX value of less than 2, nor more than the
;	'window' size (MINBUF*16) specified via the MINBUF assembly time
;	variable.
;  DX:	A copy of the last CX value as returned immediately after the
;	last call to 'stdrw.' Unless CX equals this value, 'rfshchk'
;	assume that CX might be out of date.
;Outputs:
;  carry set: CX is less than BX, but CX is potentially out of date. A
;	call to 'stdrw' is in order, to possibly fill/empty buffers, and
;	to get an accurate CX value. In this case, 'rfshchk' has already
;	decremented SI by one, so that the current byte can be refetched
;	after the 'stdrw' call.
;  zero set (no carry): CX is up to date, and is less than BX,
;	indicating that one cannot process BX bytes without further
;	consideration of counts. This occurs only as one approaches the
;	end of the input stream.
;  BX:	(no carry case) One less than BX was on input, but also no more
;	than CX - 1. You can always reduce CX by BX after returning from
;	'rfshchk' without fear of reducing CX to or past zero, so that
;	you can continue the loop with a 'loop' instruction.
;
rfshchk proc	near
	cmp	cx,bx	;plenty of bytes?
	jb	rfsh02	;jump if not
	dec	bx
	test	sp,sp	;clear carry, reset zero flag
	ret
;
rfsh02: cmp	cx,dx	;count up to date?
	jnb	rfsh04	;jmp if so, no r/w update rqd
	dec	si	;prepare to refresh count and refetch byte
	ret		;return with carry, bfrs/cnt need refresh
;
rfsh04: mov	bx,cx
	dec	bx
	cmp	sp,sp	;clear carry, set zero flag
	ret
rfshchk endp
;
;'stdrw' computes the lesser of the input data available and output
;space available, returning the value in CX. It calls read and write as
;required if either value is less than MINBUF*16, except that the input
;data available will drop below MINBUF*16 as the end of the input file
;is reached. 'stdrw' also clears AX and BX and copies CX to DX
;
stdrwx	proc	near
doaread:xchg	bx,cx	;BX=bufmax, CX=remaining cnt, DX=bufmin
	call	read
stdrw:	mov	bx,cs:inbufend
	mov	dx,cs:bufmin
	sub	bx,si		;bytes remaining in inbuf
	mov	cx,cs:bufmax
	cmp	dx,bx
	jb	std02		;jmp if bufmin less than remaining
;remaining bytes less than or equal to bufmin. we can do a read
	cmp	cx,cs:inbufend
	je	doaread
std02:	sub	cx,di		;compute space remaining (incl guard)
	cmp	dx,cx		;cmpr minbuf with space remaining
	jnb	writenow	;jmp if space remaining <= minbuf
	cmp	cx,bx		;output space remain w/input byte avail
	jb	cxok		;jmp if output space lower
	mov	cx,bx		;CX=min(in bytes avail,out space avail)
cxok:	mov	dx,cx
	xor	ax,ax
	xor	bx,bx
	ret
;
writenow:call	write
	jmp	short stdrw
stdrwx	endp
;
	assume	ds:_TEXT
close	proc	near
	mov	cx,-1
	mov	bx,infile
	inc	bx
	jz	clos02
	dec	bx
	jz	clos02
	mov	ah,3Eh
	int	21h
	mov	infile,cx
clos02: mov	bx,outfile
	inc	bx
	jz	clos04
	dec	bx
	dec	bx
	jz	clos04
	inc	bx
	mov	ah,3Eh
	int	21h
	mov	outfile,cx
clos04: ret
close	endp
;
putstderr proc	near
;In:  DS:DX points to a '$' delimited message for stderr
;Out: AX destroyed, direction flag as returned by DOS
	push	es
	push	ds
	pop	es
	push	di
	mov	di,dx
	mov	ax,(40h*256)+"$" ;AH=40: write to handle
	push	cx
	mov	cx,1001h
	cld
	repne	scasb	;scan for '$'
	neg	cx	;set length, low three nibbles
	and	ch,0Fh	;clear high length nibble
	push	bx
	mov	bx,2	;stderror handle
	int	21h	;write to stderror
	pop	bx
	pop	cx
	pop	di
	pop	es
	ret
putstderr endp
;
emsg	proc	near
	test	ch,ch
	jz	emsg02
	call	putstderr
	mov	cx,1
emsg02: ret
emsg	endp
;
errexit:pop	dx
	push	cs
	pop	ds
	call	close
	call	putstderr
errprompt:call	err02
	db	'Press any key ...',13,10,'$'
err02:	pop	dx
	call	putstderr
keywait:mov	ah,1
	int	16h	;reset z flag if bytes stacked in kybd buffer
	pushf
	mov	ah,0
	int	16h	;get a keyboard byte
	popf
	jnz	keywait 	;loop to clear stacked input bytes
	mov	ax,4C01h
	int	21h		;exit return code 1
;
;'write' writes output data and updates ES:DI to an available buffer
;location. It should not be called unless DI>=(bufmax-bufmin), except
;for the final call prior to close
;
;Inputs:
;  outfile: file handle
;  ES:DI: current output buffer target location
;Outputs:
;  ES:DI: updated to new target location
;
	assume	ds:nothing
write	proc	near
	push	ds		;save DS
	push	si		;save SI
	push	es
	pop	ds		;DS to write segment
	mov	ah,40h		;write
	mov	bx,cs:outfile
	mov	cx,cs:bufmax
	sub	cx,cs:bufmin	;CX=normal write length
	mov	si,cx		;start address of portion to move down
	cmp	cx,di		;more than what is actually there?
	jb	writ02		;jump if not
	mov	cx,di		;restrict write to actual
writ02: xor	dx,dx
	jcxz	writ04		;jump if no data to write
	dowrite
	int	21h
	cld
	jc	werror
	cmp	ax,cx
	jb	werror
writ04: mov	cx,di		;end address of portion to move down
	xor	di,di
	sub	cx,si		;length of portion to move
	jb	writ06		;skip move if start addr>end address
	rep	movsb
writ06: pop	si		;restore SI
	pop	ds		;restore DS
	ret
;
werror: call	errexit
	db	7,'Write error on output file',13,10,'$'
;
write	endp
;
done:	call	write
	call	write		;in case there was some in guard area
	push	cs
	pop	ds
	assume	ds:_TEXT
	mov	ax,5701h
	mov	bx,infile
	test	bx,bx
	jz	done02
	mov	bx,outfile
	cmp	bx,1
	je	done02
	mov	cx,datecx
	mov	dx,datedx
	int	21h
	jc	werror
done02: call	close
	xor	cx,cx
	call	warnings	;let includer issue warnings
	jcxz	exit
	jmp	errprompt
exit:	mov	ax,4C00h
	int	21h		;exit retcode 0
;
	assume	ds:nothing
